/*
 * Copyright (c) 2023, 2024, Oracle and/or its affiliates.
 *
 * Licensed under the Universal Permissive License v 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 */

#include <stddef.h>
#include <string.h>

#include "oci.h"
#include "utils.h"

#define MEMBER(m)	offsetof(MAPPING, m)
#define MEMBER_END	0, NULL, 0


#define MAPPING struct oci_os_object_summary

struct json_mapping oci_os_object_summary_mapping[] = {
	{ MEMBER(name),	"name", TYPE_STRING },
	{ MEMBER(size), "size", TYPE_NUMBER },
	{ MEMBER(etag),	"etag", TYPE_STRING },
	{ MEMBER(created), "timeCreated", TYPE_DATE },
	{ MEMBER(modified), "timeModified", TYPE_DATE },
	{ MEMBER_END }
};

#undef MAPPING

#define MAPPING struct oci_os_rename_object_details

struct json_mapping oci_os_rename_object_details_mapping[] = {
	{ MEMBER(new_name), "newName", TYPE_STRING },
	{ MEMBER(new_obj_if_match_etag), "newObjIfMatchETag", TYPE_STRING },
	{ MEMBER(new_obj_if_none_match_etag),
	  "newObjIfNoneMatchETag", TYPE_STRING },
	{ MEMBER(source_name), "sourceName", TYPE_STRING },
	{ MEMBER(src_obj_if_match_etag), "srcObjIfMatchETag", TYPE_STRING },
	{ MEMBER_END }
};

#undef MAPPING

/*
 * MultipartUpload Mapping
 */

struct oci_os_create_mpu_details {
	char *object;
	char *cache_control;
	char *content_disposition;
	char *content_encoding;
	char *content_language;
	char *content_type;
	char *storage_tier;
};

#define MAPPING struct oci_os_create_mpu_details

struct json_mapping oci_os_create_mpu_details_mapping[] = {
	{ MEMBER(object), "object", TYPE_STRING },
	{ MEMBER(cache_control), "cache_control", TYPE_STRING },
	{ MEMBER(content_disposition), "content_disposition", TYPE_STRING },
	{ MEMBER(content_encoding), "content_encoding", TYPE_STRING },
	{ MEMBER(content_language), "content_language", TYPE_STRING },
	{ MEMBER(content_type), "content_type", TYPE_STRING },
	{ MEMBER(storage_tier), "storage_tier", TYPE_STRING },
	{ MEMBER_END }
};

#undef MAPPING

#define MAPPING struct oci_os_mpu

struct json_mapping oci_os_mpu_mapping[] = {
	{ MEMBER(bucket), "bucket", TYPE_STRING },
	{ MEMBER(namespace), "namespace", TYPE_STRING },
	{ MEMBER(object), "object", TYPE_STRING },
	{ MEMBER(storage_tier), "storageTier", TYPE_STRING },
	{ MEMBER(time_created), "timeCreated", TYPE_DATE },
	{ MEMBER(upload_id), "uploadId", TYPE_STRING },
	{ MEMBER_END }
};

#undef MAPPING

#define MAPPING struct oci_os_mpu_part_summary

struct json_mapping oci_os_mpu_part_summary_mapping[] = {
	{ MEMBER(part_number), "partNumber", TYPE_NUMBER },
	{ MEMBER(size), "size", TYPE_NUMBER },
	{ MEMBER(etag), "etag", TYPE_STRING },
	{ MEMBER(md5), "md5", TYPE_STRING },
	{ MEMBER_END }
};

#undef MAPPING

/*
 * URL paths for the Object Storage API: /n/<namespace>/b/<bucket>/o/<object>
 */

#define OCI_OS_ROOT_NAMESPACE		"/n"
#define OCI_OS_PATH_NAMESPACE		OCI_OS_ROOT_NAMESPACE "/%s"
#define OCI_OS_ROOT_BUCKET		OCI_OS_PATH_NAMESPACE "/b"
#define OCI_OS_PATH_BUCKET		OCI_OS_ROOT_BUCKET "/%s"
#define OCI_OS_ROOT_OBJECT		OCI_OS_PATH_BUCKET "/o"
#define OCI_OS_PATH_OBJECT		OCI_OS_ROOT_OBJECT "/%s"
#define OCI_OS_ROOT_ACTIONS 		OCI_OS_PATH_BUCKET "/actions"
#define OCI_OS_PATH_ACTIONS 		OCI_OS_ROOT_ACTIONS "/%s"
#define OCI_OS_ROOT_UPLOAD 		OCI_OS_PATH_BUCKET "/u"
#define OCI_OS_PATH_UPLOAD 		OCI_OS_ROOT_UPLOAD "/%s"

#define OCI_OS_ACTION_RENAME_OBJECT	"renameObject"

#define PATH_NAMESPACE(config)				\
	(config)->os_namespace

#define PATH_BUCKET(config)				\
	PATH_NAMESPACE(config), (config)->os_bucket

static char *oci_os_path_bucket(struct oci_config *config,
				const char *bucket_name)
{
	return strfmt(OCI_OS_PATH_BUCKET, PATH_NAMESPACE(config), bucket_name);
}

static char *oci_os_path_object(struct oci_config *config,
				const char *object_name)
{
	char *escaped_name;
	char *path;

	escaped_name = escape_path(object_name);
	path = strfmt(OCI_OS_PATH_OBJECT, PATH_BUCKET(config), escaped_name);
	if (escaped_name != object_name)
		free(escaped_name);

	return path;
}

int oci_os_head_bucket(struct oci_config *config, const char *bucket_name,
		       struct oci_os_bucket_head *bucket,
		       struct oci_error *error)
{
	struct oci_header_field response_header[] = { { "ETag", NULL } };
	char *path;
	int err;

	path = oci_os_path_bucket(config, bucket_name);
	if (!path) {
		oci_error(error, "Failed to create bucket path");
		return -1;
	}

	err = oci_request_head(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
			       OCI_HEADERS_RESPONSE_ARRAY(response_header),
			       error);
	free(path);
	if (err)
		return err;

	if (!response_header[0].value)
		return -1;

	bucket->name = (char *)bucket_name;
	bucket->etag = response_header[0].value;

	return 0;
}

void oci_os_bucket_head_fini(struct oci_os_bucket_head *bucket)
{
	if (!bucket)
		return;

	free(bucket->etag);
}

int oci_os_head_object(struct oci_config *config,
		       const char *object_name,
		       struct oci_os_object_head *object,
		       struct oci_error *error)
{
	struct oci_header_field response_header[] = { { "ETag", NULL },
						      { "content-length", NULL },
						      { "last-modified", NULL } };
	char *content_length;
	char *last_modified;
	int rv = -1;
	char *etag;
	char *path;
	char *s;
	int err;

	path = oci_os_path_object(config, object_name);
	if (!path) {
		oci_error(error, "Failed to create object path");
		return -1;
	}

	err = oci_request_head(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
			       OCI_HEADERS_RESPONSE_ARRAY(response_header),
			       error);
	free(path);
	if (err)
		return err;

	etag = response_header[0].value;
	content_length = response_header[1].value;
	last_modified = response_header[2].value;

	if (!etag || !content_length || !last_modified)
		goto done;

	object->content_length = strtol(content_length, &s, 10);
	if (*s != '\0')
		goto done;

	s = strptime(last_modified, "%a, %d %b %Y %T %Z",
		     &object->last_modified);
	if (*s != '\0')
		goto done;

	object->name = (char *)object_name;
	object->etag = etag;

	/* etag reference is borrowed by object->etag, do not free it */
	etag = NULL;
	rv = 0;

done:
	free(etag);
	free(content_length);
	free(last_modified);

	return rv;
}

void oci_os_object_head_fini(struct oci_os_object_head *object)
{
	if (!object)
		return;

	free(object->etag);
}

int oci_os_list_object(struct oci_config *config,
		       const char *object_name,
		       struct oci_os_object_summary *object,
		       struct oci_error *error)
{
	struct oci_os_list_objects_param param = {0};
	struct oci_os_list_objects list_objects;
	int count;

	/*
	 * There is no parameter to directly query a single object. But
	 * the 'start' parameter queries object with a name >= start,
	 * and 'end' queries object with a name < end.
	 *
	 * So to get a specific object, we use start=<name> and
	 * end=<name> + '\1'.
	 */

	param.fields = "size";
	param.start = (char *)object_name;
	param.end = strfmt("%s\1", object_name);
	if (!param.end)
		return -1;

	count = oci_os_list_objects(config, &list_objects, &param, error);
	free(param.end);

	if (count < 0) {
		/* error */
		return -1;
	}

	if (count == 0) {
		/* object doesn't exist */
		*object = (struct oci_os_object_summary) { 0 };
		return 0;
	}

	if (count > 1) {
		/* unexpected! we shouldn't get multiple objects */
		oci_os_list_objects_fini(&list_objects);
		return -1;
	}

	*object = list_objects.objects[0];

	/*
	 * We will return the single object we got. So, before clearing
	 * list_objects, set 'objects_count' to 0 (so that the object
	 * content is not clear), but do not clear 'objects' so it gets
	 * freed.
	 */
	list_objects.objects_count = 0;
	oci_os_list_objects_fini(&list_objects);

	return 1;
}

void oci_os_object_summary_fini(struct oci_os_object_summary *object)
{
	if (!object)
		return;

	free(object->name);
	free(object->etag);
}

void oci_os_list_objects_fini(struct oci_os_list_objects *list_objects)
{
	struct oci_os_object_summary *objects;
	char **prefixes;
	int i;

	if (!list_objects)
		return;

	free(list_objects->next_start_with);

	objects = list_objects->objects;
	for (i = 0; i < list_objects->objects_count; i++)
		oci_os_object_summary_fini(&objects[i]);

	prefixes = list_objects->prefixes;
	for (i = 0; i < list_objects->prefixes_count; i++)
		free(prefixes[i]);

	free(objects);
	free(prefixes);
}

static int oci_parse_prefixes(json_t *result,
			      struct oci_os_list_objects *list_objects)
{
	json_t *json_prefix_list;
	char ***prefixes;
	int count;

	json_prefix_list = json_object_get(result, "prefixes");
	if (!json_prefix_list) {
		if (list_objects) {
			list_objects->prefixes = NULL;
			list_objects->prefixes_count = 0;
		}
		return 0;
	}

	prefixes = list_objects ? &list_objects->prefixes : NULL;

	count = json_array_to_str_array(json_prefix_list, prefixes);
	if (count < 0)
		return -1;

	if (list_objects)
		list_objects->prefixes_count = count;

	return count;
}

static int oci_parse_objects(json_t *result,
			     struct oci_os_list_objects *list_objects)
{
	json_t *json_obj_list;
	void **objects;
	int count;

	json_obj_list = json_object_get(result, "objects");
	if (!json_obj_list)
		return -1;

	objects = list_objects ? (void **)&list_objects->objects : NULL;

	count = json_array_to_struct_array(json_obj_list, objects,
					   sizeof(struct oci_os_object_summary),
					   oci_os_object_summary_mapping);
	if (count < 0)
		return -1;

	if (list_objects)
		list_objects->objects_count = count;

	return count;
}

static int oci_parse_next_start_with(json_t *result,
				     struct oci_os_list_objects *list_objects,
				     char **nextp)
{
	json_t *json_str;
	char *next;

	if (!list_objects && !nextp)
		return 0;

	json_str = json_object_get(result, "nextStartWith");
	if (!json_str) {
		next = NULL;
		goto done;
	}

	next = strdup(json_string_value(json_str));
	if (!next)
		return -1;

done:
	if (list_objects)
		list_objects->next_start_with = next;
	if (nextp)
		*nextp = next;

	return 0;
}

/*
 * Add pagination parameters to the params array; pcount should be the
 * number of parameters already present in the array.
 *
 * Return the updated number of parameters in the array.
 */
static int param_add_pagination(struct parameter *params, int pcount,
				struct oci_list_pagination *pagination)
{
	if (pagination) {
		PARAM_STR_SET(&params[pcount], "page", pagination->page);
		pcount++;
		PARAM_INT_SET(&params[pcount], "limit", pagination->limit);
		pcount++;
	}

	return pcount;
}

static char *oci_os_path_list_objects(struct oci_config *config,
				      struct oci_os_list_objects_param *param)
{
	char *path;

	if (!param)
		return strfmt(OCI_OS_ROOT_OBJECT, PATH_BUCKET(config));

	struct parameter params[] = { PARAM_STR("prefix", param->prefix),
				      PARAM_STR("start", param->start),
				      PARAM_STR("end", param->end),
				      PARAM_STR("delimiter", param->delimiter),
				      PARAM_STR("fields", param->fields),
				      PARAM_INT("limit", param->limit) };


	path = create_path_query(OCI_OS_ROOT_OBJECT,
				 params, ARRAY_SIZE(params),
				 PATH_BUCKET(config));

	return path;
}

static int oci_os_list_objects_page(struct oci_config *config,
				    struct oci_os_list_objects *list_objects,
				    struct oci_os_list_objects_param *param,
				    char **nextp, struct oci_error *error)
{
	json_t *result;
	char *path;
	int count;
	int err;

	path = oci_os_path_list_objects(config, param);
	if (!path) {
		oci_error(error, "Failed create list objects path");
		return -1;
	}

	result = oci_request_get_json(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
				      NULL, error);
	free(path);
	if (!result)
		return -1;

	if (list_objects)
		*list_objects = (struct oci_os_list_objects){ 0 };

	/*
	 * Parse prefixes if they were requested.
	 */
	if (param && param->delimiter) {
		count = oci_parse_prefixes(result, list_objects);
		if (count < 0) {
			oci_error(error, "Failed to parse OCI prefixes");
			json_decref(result);
			return -1;
		}
	}

	/*
	 * Parse NextStartWith.
	 */
	err = oci_parse_next_start_with(result, list_objects, nextp);
	if (err) {
		oci_error(error, "Failed to parse OCI NextStartWith");
		json_decref(result);
		/* clear list_objects to free prefixes */
		oci_os_list_objects_fini(list_objects);
		return -1;
	}

	/*
	 * Parse objects.
	 */
	count = oci_parse_objects(result, list_objects);
	if (count < 0) {
		oci_error(error, "Failed to parse list objects");
		json_decref(result);
		/* clear list_objects to free prefixes */
		oci_os_list_objects_fini(list_objects);
		return -1;
	}

	json_decref(result);

	return count;
}

/*
 * List objects.
 *
 * If list_objects is not NULL then return objects for the parameters
 * specified in 'param'. Because of paging and the 'limit' attribute,
 * the list can be incomplete. In that case, the 'next_start_with'
 * attribute will be set in 'param', and the function returns the number
 * of objects returned in 'list_objects'.
 *
 * If list_objects is NULL then the function returns the total number of
 * objects for the parameters specified in 'param'.
 */
int oci_os_list_objects(struct oci_config *config,
			struct oci_os_list_objects *list_objects,
			struct oci_os_list_objects_param *param,
			struct oci_error *error)
{
	struct oci_os_list_objects_param _param = { 0 };
	char *next = NULL;
	char *start;
	int limit;
	int count;
	int c;

	c = oci_os_list_objects_page(config, list_objects, param, &next,
				     error);
	if (c < 0)
		return c;

	if (list_objects) {
		/*
		 * Do not free 'next' because it is also referenced
		 * in list_objects.
		 */
		return c;
	}

	/*
	 * If list_objects is NULL then count the total number of
	 * objects and return that value.
	 */

	if (!param)
		param = &_param;

	start = param->start;
	limit = param->limit;
	count = c;

	while (next && (!limit || count < limit)) {
		param->start = next;
		c = oci_os_list_objects_page(config, NULL, param, &next,
					     error);
		free(param->start);
		if (c < 0) {
			count = c;
			break;
		}
		count += c;
	}

	free(next);
	param->start = start;

	return count;
}

static int get_object_headers(struct oci_os_get_object_param *param,
			      char **header_p, int *header_count_p)
{
	off_t range_begin, range_end;
	int header_count;
	char *header;

	if (!param) {
		*header_p = NULL;
		*header_count_p = 0;
		return 0;
	}

	if (param->range_start < 0 || param->range_size < 0)
		return -1;

	range_begin = param->range_start;
	range_end = range_begin + param->range_size - 1;

	header = strfmt("range: bytes=%zu-%zu", range_begin, range_end);
	if (!header)
		return -1;

	header_count = 1;

	*header_p = header;
	*header_count_p = header_count;

	return 0;
}

static char *oci_os_get_object_prepare(struct oci_config *config,
				       const char *object_name,
				       struct oci_os_get_object_param *param,
				       char **header_p, int *header_count_p,
				       struct oci_error *error)
{
	char *path;
	int err;

	path = oci_os_path_object(config, object_name);
	if (!path) {
		oci_error(error, "Failed to create object path");
		return NULL;
	}

	err = get_object_headers(param, header_p, header_count_p);
	if (err) {
		oci_error(error, "Failed to get object headers");
		free(path);
		return NULL;
	}

	return path;
}

/*
 * Get an object (or a part of an object if param specifies a range_start
 * and a range_size) and store it in the specified buffer and size.
 */
int oci_os_get_object(struct oci_config *config, const char *object_name,
		      char **bufp, size_t *sizep,
		      struct oci_os_get_object_param *param,
		      struct oci_error *error)
{
	int header_count, err;
	char *path, *header;

	path = oci_os_get_object_prepare(config, object_name, param,
					 &header, &header_count, error);
	if (!path)
		return -1;

	err = oci_request_get(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
			      bufp, sizep,
			      OCI_HEADERS_REQUEST(&header, header_count),
			      error);

	free(header);
	free(path);

	return err;
}

/*
 * Get an object (or a part of an object if param specifies a range_start
 * and a range_size) and store it in a file at the specified file offset
 * and size.
 */
int oci_os_get_object_to_file(struct oci_config *config,
			      const char *object_name,
			      int fd, off_t offset, size_t size,
			      struct oci_os_get_object_param *param,
			      struct oci_error *error)
{
	int header_count, err;
	char *path, *header;

	path = oci_os_get_object_prepare(config, object_name, param,
					 &header, &header_count, error);
	if (!path)
		return -1;

	err = oci_request_get_to_file(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
				      fd, offset, size,
				      OCI_HEADERS_REQUEST(&header, header_count),
				      error);

	free(header);
	free(path);

	return err;
}

/*
 * Put an object. The object is created from the specified buffer and size.
 */
int oci_os_put_object(struct oci_config *config, const char *object_name,
		      char *buffer, size_t size, struct oci_error *error)
{
	char *path;
	int err;

	path = oci_os_path_object(config, object_name);
	if (!path) {
		oci_error(error, "Failed to create object path");
		return -1;
	}

	err = oci_request_put(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
			      buffer, size, NULL, NULL, NULL, error);

	free(path);

	return err;
}

/*
 * Put an object. The object is created from the entire content of the
 * specified file.
 */
int oci_os_put_object_from_file(struct oci_config *config,
				const char *object_name, int fd,
				struct oci_error *error)
{
	char *path;
	int err;

	path = oci_os_path_object(config, object_name);
	if (!path) {
		oci_error(error, "Failed to create object path");
		return -1;
	}

	err = oci_request_put_from_file(config,
					OCI_ENDPOINT_OBJECT_STORAGE, path,
					fd, 0, 0, NULL, NULL, NULL, error);

	free(path);

	return err;
}

int oci_os_delete_object(struct oci_config *config, const char *object_name,
			 struct oci_error *error)
{
	char *path;
	int err;

	path = oci_os_path_object(config, object_name);
	if (!path) {
		oci_error(error, "Failed to create object path");
		return -1;
	}

	err = oci_request_delete(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
				 error);

	free(path);

	return err;
}

int oci_os_rename_object(struct oci_config *config, const char *old_name,
			 const char *new_name, struct oci_error *error)
{
	struct oci_os_rename_object_details rename = { 0 };
	json_t *object;
	char *body;
	char *path;
	int err;

	path = strfmt(OCI_OS_PATH_ACTIONS, PATH_BUCKET(config),
		      OCI_OS_ACTION_RENAME_OBJECT);
	if (!path) {
		oci_error(error, "Failed to create object path");
		return -1;
	}

	rename.source_name = old_name;
	rename.new_name = new_name;

	object = json_struct_to_object(&rename,
				       oci_os_rename_object_details_mapping);
	if (!object) {
		oci_error(error, "Failed to create JSON request");
		free(path);
		return -1;
	}

	body = json_dumps(object, 0);
	if (!body) {
		oci_error(error, "Failed to create request body");
		free(path);
		json_decref(object);
		return -1;
	}

	err = oci_request_post(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
			       body, strlen(body), NULL, error);

	free(body);
	json_decref(object);
	free(path);

	return err;
}

char *oci_os_get_namespace(struct oci_config *config, struct oci_error *error)
{
	char *namespace = NULL;
	json_t *result;

	result = oci_request_get_json(config,
				      OCI_ENDPOINT_OBJECT_STORAGE,
				      OCI_OS_ROOT_NAMESPACE,
				      NULL, error);
	if (!result)
		return NULL;

	if (!json_is_string(result)) {
		oci_error(error, "Response is not a JSON string");
		goto done;
	}

	namespace = (char *)json_string_value(result);
	if (!namespace)
		goto done;

	namespace = strdup(namespace);

done:
	json_decref(result);
	return namespace;
}

/*
 * MultipartUpload
 */

static char *oci_os_root_upload(struct oci_config *config,
				struct oci_list_pagination *pagination)
{
	struct parameter params[2];
	int pcount = 0;
	char *path;

	if (!pagination)
		return strfmt(OCI_OS_ROOT_UPLOAD, PATH_BUCKET(config));

	pcount = param_add_pagination(params, pcount, pagination);

	path = create_path_query(OCI_OS_ROOT_UPLOAD, params, ARRAY_SIZE(params),
				 PATH_BUCKET(config));

	return path;
}

static char *oci_os_path_upload(struct oci_config *config,
				const char *object_name,
				const char *upload_id, const int upload_part,
				struct oci_list_pagination *pagination)
{
	struct parameter params[4] = { PARAM_STR("uploadid", upload_id),
				       PARAM_INT("uploadpartnum", upload_part) };
	char *escaped_name;
	int pcount = 2;
	char *path;

	escaped_name = escape_path(object_name);
	if (!escaped_name)
		return NULL;

	pcount = param_add_pagination(params, pcount, pagination);

	path = create_path_query(OCI_OS_PATH_UPLOAD, params, pcount,
				 PATH_BUCKET(config), escaped_name);

	if (escaped_name != object_name)
		free(escaped_name);

	return path;
}

/* MultipartUpload Resource */

void oci_os_mpu_fini(struct oci_os_mpu *upload)
{
	if (!upload)
		return;

	free(upload->bucket);
	free(upload->namespace);
	free(upload->object);
	free(upload->storage_tier);
	free(upload->upload_id);
}

/* CreateMultipartUpload */

int oci_os_create_multipart_upload(struct oci_config *config,
				   const char *object_name,
				   struct oci_os_mpu *upload,
				   struct oci_error *error)
{
	struct oci_os_create_mpu_details upload_details = { 0 };
	json_t *object;
	json_t *result;
	char *body;
	char *path;
	int err;

	path = oci_os_root_upload(config, NULL);
	if (!path) {
		oci_error(error, "Failed create root upload path");
		return -1;
	}

	upload_details.object = (char *)object_name;

	object = json_struct_to_object(&upload_details,
				       oci_os_create_mpu_details_mapping);
	if (!object) {
		oci_error(error, "Failed to create JSON request");
		free(path);
		return -1;
	}

	body = json_dumps(object, 0);
	if (!body) {
		oci_error(error, "Failed to create request body");
		free(path);
		json_decref(object);
		return -1;
	}

	result = oci_request_post_json(config, OCI_ENDPOINT_OBJECT_STORAGE,
				       path, body, strlen(body), NULL, error);

	free(body);
	json_decref(object);
	free(path);

	if (!result)
		return -1;

	if (upload) {
		err = json_object_to_struct(result, upload,
					    oci_os_mpu_mapping);
	} else {
		err = 0;
	}

	json_decref(result);

	return err;
}

/* UploadPart */

int oci_os_upload_part(struct oci_config *config, const char *object_name,
		       const char *upload_id, const int upload_part,
		       char *buffer, size_t size,
		       oci_request_progress_f progress, void *progress_data,
		       char **etag_p, struct oci_error *error)
{
	struct oci_header_field response_header[] = { { "ETag", NULL } };
	char *path;
	int err;

	path = oci_os_path_upload(config, object_name, upload_id, upload_part,
				  NULL);
	if (!path) {
		oci_error(error, "Failed to create upload path");
		return -1;
	}

	err = oci_request_put(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
			      buffer, size,
			      OCI_HEADERS_RESPONSE_ARRAY(response_header),
			      progress, progress_data, error);

	free(path);
	if (err)
		return err;

	if (etag_p)
		*etag_p = response_header[0].value;

	return 0;
}

int oci_os_upload_part_from_file(struct oci_config *config,
				 const char *object_name,
				 const char *upload_id, const int upload_part,
				 int fd, off_t offset, size_t size,
				 oci_request_progress_f progress,
				 void *progress_data,
				 char **etag_p, struct oci_error *error)
{
	struct oci_header_field response_header[] = { { "ETag", NULL } };
	char *path;
	int err;

	if (!size) {
		*etag_p = NULL;
		return 0;
	}

	path = oci_os_path_upload(config, object_name, upload_id, upload_part,
		NULL);
	if (!path) {
		oci_error(error, "Failed to create upload path");
		return -1;
	}

	err = oci_request_put_from_file(config, OCI_ENDPOINT_OBJECT_STORAGE,
					path, fd, offset, size,
					OCI_HEADERS_RESPONSE_ARRAY(response_header),
					progress, progress_data, error);

	free(path);
	if (err)
		return err;

	if (etag_p)
		*etag_p = response_header[0].value;

	return 0;
}

/* CommitMultipartUpload */

int oci_os_commit_mpu_arg_init(struct oci_os_commit_mpu_arg *arg)
{
	int err;

	arg->details = NULL;
	arg->parts_to_commit = NULL;
	arg->parts_to_exclude = NULL;

	arg->details = json_object();
	if (!arg->details)
		goto error;

	arg->parts_to_commit = json_array();
	if (!arg->parts_to_commit)
		goto error;

	arg->parts_to_exclude = json_array();
	if (!arg->parts_to_exclude)
		goto error;

	err = json_object_set(arg->details, "partsToCommit",
			      arg->parts_to_commit);
	if (err)
		goto error;

	err = json_object_set(arg->details, "partsToExclude",
			      arg->parts_to_exclude);
	if (err)
		goto error;

	return 0;

error:
	oci_os_commit_mpu_arg_fini(arg);
	return -1;
}

void oci_os_commit_mpu_arg_fini(struct oci_os_commit_mpu_arg *arg)
{
	json_decref(arg->details);
	json_decref(arg->parts_to_commit);
	json_decref(arg->parts_to_exclude);
	arg->details = NULL;
	arg->parts_to_commit = NULL;
	arg->parts_to_exclude = NULL;
}

int oci_os_commit_mpu_add_commit(struct oci_os_commit_mpu_arg *arg,
				 int pnum, char *etag)
{
	json_t *part_detail = NULL;
	json_t *pnum_json = NULL;
	json_t *etag_json = NULL;
	int err;

	if (!etag)
		return 0;

	pnum_json = json_integer(pnum);
	if (!pnum_json)
		goto error;

	etag_json = json_string(etag);
	if (!etag_json)
		goto error;

	part_detail = json_object();
	if (!part_detail)
		return -1;

	err = json_object_set(part_detail, "etag", etag_json);
	if (err)
		goto error;

	err = json_object_set(part_detail, "partNum", pnum_json);
	if (err)
		goto error;

	err = json_array_append_new(arg->parts_to_commit, part_detail);
	if (err)
		goto error;

	json_decref(pnum_json);
	json_decref(etag_json);

	return 0;

error:
	json_decref(pnum_json);
	json_decref(etag_json);
	json_decref(part_detail);

	return -1;
}

int oci_os_commit_mpu_add_exclude(struct oci_os_commit_mpu_arg *arg, int pnum)
{
	json_t *pnum_json = NULL;
	int err;

	pnum_json = json_integer(pnum);
	if (!pnum_json)
		return -1;

	err = json_array_append_new(arg->parts_to_exclude, pnum_json);

	json_decref(pnum_json);

	return err;
}

int oci_os_commit_multipart_upload(struct oci_config *config,
				   const char *object_name,
				   const char *upload_id,
				   struct oci_os_commit_mpu_arg *arg,
				   struct oci_error *error)
{
	char *body;
	char *path;
	int err;

	if (!arg || !arg->details) {
		oci_error(error, "Invalid argument");
		return -1;
	}

	path = oci_os_path_upload(config, object_name, upload_id, 0, NULL);
	if (!path) {
		oci_error(error, "Failed to create upload path");
		return -1;
	}

	body = json_dumps(arg->details, 0);
	if (!body) {
		oci_error(error, "Failed to create request body");
		free(path);
		return -1;
	}

	err = oci_request_post(config, OCI_ENDPOINT_OBJECT_STORAGE,
			       path, body, strlen(body), NULL, error);

	free(body);
	free(path);

	return err;
}

/* AbortMultipartUpload */

int oci_os_abort_multipart_upload(struct oci_config *config,
				  const char *object_name,
				  const char *upload_id,
				  struct oci_error *error)
{
	char *path;
	int err;

	path = oci_os_path_upload(config, object_name, upload_id, 0, NULL);
	if (!path) {
		oci_error(error, "Failed to create upload path");
		return -1;
	}

	err = oci_request_delete(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
				 error);

	free(path);

	return err;
}

/* ListMultipartUploads */

void oci_os_list_mpus_fini(struct oci_os_list_mpus *list_uploads)
{
	struct oci_os_mpu *uploads;
	int i;

	if (!list_uploads)
		return;

	uploads = list_uploads->uploads;
	for (i = 0; i < list_uploads->uploads_count; i++)
		oci_os_mpu_fini(&uploads[i]);

	free(uploads);
}

int oci_os_list_multipart_uploads(struct oci_config *config,
				  struct oci_os_list_mpus *list_uploads,
				  struct oci_list_pagination *pagination,
				  struct oci_error *error)
{
	struct oci_header_field response_header[] = { { "opc-next-page", NULL } };
	json_t *result;
	void **objects;
	char *path;
	int count;

	path = oci_os_root_upload(config, pagination);
	if (!path) {
		oci_error(error, "Failed create root upload path");
		return -1;
	}

	result = oci_request_get_json(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
				      OCI_HEADERS_RESPONSE_ARRAY(response_header),
				      error);
	free(path);
	if (!result)
		return -1;

	objects = (void **)&list_uploads->uploads;

	count = json_array_to_struct_array(result, objects,
					   sizeof(struct oci_os_mpu),
					   oci_os_mpu_mapping);
	if (count < 0) {
		oci_error(error, "Failed to parse multipart uploads");
		json_decref(result);
		return -1;
	}

	json_decref(result);

	list_uploads->uploads_count = count;
	if (pagination)
		pagination->next = response_header[0].value;

	return count;

}

/* ListMultipartUploadParts */

void oci_os_mpu_part_fini(struct oci_os_mpu_part_summary *part)
{
	if (!part)
		return;

	free(part->etag);
	free(part->md5);
}

void oci_os_list_mpu_parts_fini(struct oci_os_list_mpu_parts *list_parts)
{
	struct oci_os_mpu_part_summary *parts;
	int i;

	if (!list_parts)
		return;

	parts = list_parts->parts;
	for (i = 0; i < list_parts->parts_count; i++)
		oci_os_mpu_part_fini(&parts[i]);

	free(parts);
}

int oci_os_list_multipart_upload_parts(struct oci_config *config,
				       const char *object_name,
				       const char *upload_id,
				       struct oci_os_list_mpu_parts *list_parts,
				       struct oci_list_pagination *pagination,
				       struct oci_error *error)
{
	struct oci_header_field response_header[] = { { "opc-next-page", NULL } };
	json_t *result;
	void **objects;
	char *path;
	int count;

	path = oci_os_path_upload(config, object_name, upload_id, 0,
				  pagination);
	if (!path) {
		oci_error(error, "Failed to create upload path");
		return -1;
	}

	result = oci_request_get_json(config, OCI_ENDPOINT_OBJECT_STORAGE, path,
				      OCI_HEADERS_RESPONSE_ARRAY(response_header),
				      error);
	free(path);
	if (!result)
		return -1;

	objects = (void **)&list_parts->parts;

	count = json_array_to_struct_array(result, objects,
					   sizeof(struct oci_os_mpu_part_summary),
					   oci_os_mpu_part_summary_mapping);
	if (count < 0) {
		oci_error(error, "Failed to parse multipart upload parts");
		json_decref(result);
		return -1;
	}

	json_decref(result);

	list_parts->parts_count = count;
	if (pagination)
		pagination->next = response_header[0].value;

	return count;

}

/*
 * Do an IMDS request and return the result as a NUL ('\0') terminated
 * string.
 */
char *oci_imds_get_str(struct oci_config *config, char *path,
		       struct oci_error *error)
{
	char *buf = NULL;
	size_t size;
	int err;

	err = oci_request_imds_get(config, path, &buf, &size, error);
	if (err)
		return NULL;
	buf[size] = '\0';

	return buf;
}
