/*
 * Copyright (c) 2023, 2024, Oracle and/or its affiliates.
 *
 * Licensed under the Universal Permissive License v 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 */

#include <ctype.h>
#include <errno.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>
#include <openssl/sha.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "oci.h"
#include "utils.h"

#define OCI_REQUEST_IMDS_URL	"http://169.254.169.254/opc/v2/%s"

#define OCI_REQUEST_BUFSIZE 4096

#define OCI_DEBUG_REQUEST(request, fmt, ...)				\
	OCI_DEBUG((request)->config, "REQUEST %p.%d: " fmt,		\
		  (request), (request)->attempt, ##__VA_ARGS__)

struct oci_body_cfg {
	char	*buffer;
	ssize_t	buffer_size;
	int	fd;
	off_t	fd_offset;
	size_t	fd_size;
};

struct oci_request {

	/*
	 * Attributes which can be set before calling oci_request_execute()
	 * with the oci_request_set_<attr>() functions.
	 */
	int flags;
	const char *content_type;
	struct oci_headers *headers;

	/*
	 * Attributes set with parameters of oci_request_execute().
	 */
	struct oci_config *config;
	enum http_method method;
	const char *endpoint;
	char *path;
	struct oci_body_cfg request_body_cfg;
	struct oci_body_cfg response_body_cfg;
	oci_request_progress_f progress;
	void *progress_data;

	/*
	 * Attributes used during the execution of the request which
	 * don't need to be re-initialized on retry.
	 */
	char *url;
	size_t content_length;
	char *content_sha256;
	time_t start_time;
	int attempt;

	/*
	 * Runtime attributes: attributes used during the execution of
	 * the request which must be re-initialized on retry.
	 */
	CURL *curl;
	struct curl_slist *curl_headers;
	struct dstore body_request;	    /* request body */
	struct dstore body_response;	    /* response body */
	size_t recv_len;
};

#define OCI_BODY_CFG(buf, size, f, f_offset, f_size)		\
	((struct oci_body_cfg) {				\
		.buffer = buf,					\
		.buffer_size = size, 				\
		.fd = f,					\
		.fd_offset = f_offset,				\
		.fd_size = f_size })

#define OCI_BODY_NONE			OCI_BODY_CFG(NULL, -1, -1, 0, 0)
#define OCI_BODY_FD(fd, off, sz)	OCI_BODY_CFG(NULL, 0, fd, off, sz)
#define OCI_BODY_BUFFER(buf, size)	OCI_BODY_CFG(buf, size, -1, 0, 0)
#define OCI_BODY_BUFFER_DYNAMIC		OCI_BODY_BUFFER(NULL, 1024)

#define OCI_REQUEST_SHA256		0x01
#define OCI_REQUEST_IMDS		0x02

static int oci_configure_body(struct dstore *body, struct oci_body_cfg body_cfg)
{
	int err;

	if (body_cfg.fd == -1 && body_cfg.buffer_size == -1) {
		dstore_init(body);
		return 0;
	}

	if (body_cfg.fd >= 0)
		err = dstore_wrap_fd(body, body_cfg.fd,
				     body_cfg.fd_offset, body_cfg.fd_size);
	else
		err = dstore_wrap_buffer(body, body_cfg.buffer,
					 body_cfg.buffer_size);

	if (err) {
		return -1;
	}

	return 0;
}

static int oci_request_runtime_init(struct oci_request *request)
{
	int err;

	request->curl = NULL;
	request->curl_headers = NULL;
	request->body_request = (struct dstore){ 0 };
	request->body_response = (struct dstore){ 0 };
	request->recv_len = 0;

	request->curl = curl_easy_init();
	if (!request->curl)
		return -1;

	err = oci_configure_body(&request->body_request,
				 request->request_body_cfg);
	if (err)
		return -1;

	err = oci_configure_body(&request->body_response,
				 request->response_body_cfg);
	if (err)
		return -1;

	return 0;
}

static void oci_request_runtime_fini(struct oci_request *request)
{
	curl_easy_cleanup(request->curl);
	curl_slist_free_all(request->curl_headers);
	dstore_clear(&request->body_request);
	dstore_clear(&request->body_response);
}

static void oci_request_fini(struct oci_request *request)
{
	if (!request)
		return;

	oci_request_runtime_fini(request);

	free(request->url);
	free(request->content_sha256);
}

static int oci_request_prepare(struct oci_request *request,
			       struct oci_error *error)
{
	unsigned char digest[SHA256_DIGEST_LENGTH];
	int err;

	request->url = NULL;
	request->content_sha256 = NULL;

	/*
	 * Runtime (actually dstore) should be initialized before we
	 * can set content_length and content_sha256.
	 */
	err = oci_request_runtime_init(request);
	if (err) {
		oci_error(error, "Failed to init runtime");
		return -1;
	}

	request->content_length = dstore_size(&request->body_request);

	if (request->flags & OCI_REQUEST_SHA256) {
		err = dstore_sha256(&request->body_request, digest);
		if (err) {
			oci_error(error, "Failed to get dstore SHA256");
			return -1;
		}
		request->content_sha256 = base64_encode(digest,
							SHA256_DIGEST_LENGTH);
		if (!request->content_sha256) {
			oci_error(error, "Failed to encode store SHA256");
			return -1;
		}
	}

	if (request->flags & OCI_REQUEST_IMDS) {
		request->url = strfmt(OCI_REQUEST_IMDS_URL, request->path);
	} else {
		request->url = strfmt("https://%s.%s%s", request->endpoint,
				      request->config->domain, request->path);
	}

	if (!request->url) {
		oci_error(error, "Failed to build request URL");
		return -1;
	}

	return 0;
}

static void oci_request_set_flags(struct oci_request *request, int flags)
{
	request->flags = flags;
}

static void oci_request_set_content_type(struct oci_request *request,
					 const char *content_type)
{
	request->content_type = content_type;
}

static void oci_request_set_headers(struct oci_request *request,
				    struct oci_headers *headers)
{
	int i;

	request->headers = headers;
	if (!headers)
		return;

	if (headers->request_hcount == 0 || !headers->request_headers) {
		headers->request_hcount = 0;
		headers->request_headers = NULL;
	}

	if (headers->response_hcount == 0 || !headers->response_headers) {
		headers->response_hcount = 0;
		headers->response_headers = NULL;
	} else {
		for (i = 0; i < headers->response_hcount; i++)
			headers->response_headers[i].value = NULL;
	}
}

static size_t oci_request_read_cb(char *ptr, size_t size, size_t nmemb, void *data)
{
	struct oci_request *request = data;
	ssize_t send_len;
	size_t len;

	OCI_DEBUG_REQUEST(request, "READ CALLBACK: %zd*%zd bytes\n",
			  size, nmemb);

	len = size * nmemb;

	send_len = dstore_read(&request->body_request, ptr, len);
	if (send_len < 0)
		return CURL_READFUNC_ABORT;

	return send_len;
}

static size_t oci_request_write_cb(char *ptr, size_t size, size_t nmemb, void *data)
{
	struct oci_request *request = data;
	size_t recv_len, len;

	OCI_DEBUG_REQUEST(request, "WRITE CALLBACK: %zd*%zd bytes\n",
			  size, nmemb);

	len = size * nmemb;
	recv_len = request->recv_len;

	recv_len = dstore_write(&request->body_response, ptr, len);
	if (recv_len < 0)
		return 0;

	request->recv_len += recv_len;

	return recv_len;
}

static size_t oci_request_header_cb(char *buffer, size_t size, size_t nmemb,
				    void *data)
{
	struct oci_request *request = data;
	char *old_value, *new_value;
	struct oci_header_field *header;
	int i, j, nl;
	char *name;
	size_t len;

	len = size * nmemb;

	/*
	 * When debugging the request, do not print trailing newlines
	 * to avoid printing blank lines.
	 */
	nl = 0;
	for (i = len - 1; i > 0 && buffer[i] == '\n'; i--)
		nl++;

	OCI_DEBUG_REQUEST(request, "HEADER CALLBACK: %zd*%zd bytes: %.*s\n",
			  size, nmemb, (int)(len - nl), buffer);

	if (!request->headers)
		return 0;

	for (i = 0; i < request->headers->response_hcount; i++) {
		header = &request->headers->response_headers[i];
		name = header->name;
		j = strlen(name);
		if (j + 1 >= len) {
			/*
			 * The header we are looking for has a name
			 * (+ the colon ':' separator) which is too
			 * long for this buffer.
			 */
			continue;
		}
		if (buffer[j] != ':')
			continue;

		if (strncasecmp(buffer, name, j) != 0)
			continue;

		/*
		 * The header name matches, get its value (we ignore
		 * whitespaces before and after the value).
		 */
		for (j++; j < len; j++) {
			if (!isspace(buffer[j]))
				break;
		}

		while (len > j && isspace(buffer[len - 1]))
			len--;

		old_value = header->value;
		if (old_value)
			new_value = strfmt("%s,%.*s", old_value,
					   ((int)len) - j, buffer + j);
		else
			new_value = strndup(buffer + j, len - j);

		if (!new_value)
			return 0;

		free(old_value);
		header->value = new_value;
		break;
	}

	return size * nmemb;
}

/*
 * Add an header and return the new CURL header list. The header is
 * not added if it is NULL. If there is an error then the header
 * list is freed.
 */
static int add_header_const(struct curl_slist **curl_headersp, char *header)
{
	struct curl_slist *slist = *curl_headersp;
	struct curl_slist *temp;

	if (!header)
		return 0;

	temp = curl_slist_append(slist, header);
	if (!temp) {
		curl_slist_free_all(slist);
		return -1;
	}

	*curl_headersp = temp;
	return 0;
}

/*
 * Add a dynamically allocated header and return the new CURL header
 * list. Return an error if the header is NULL or if adding the header
 * fails. If there is an error then the header list is freed. In all
 * cases, the header is freed.
 */
static int add_header(struct curl_slist **curl_headersp, char *header)
{
	struct curl_slist *slist = *curl_headersp;
	struct curl_slist *temp;

	if (!header) {
		curl_slist_free_all(slist);
		return -1;
	}

	temp = curl_slist_append(slist, header);
	free(header);
	if (!temp) {
		curl_slist_free_all(slist);
		return -1;
	}

	*curl_headersp = temp;
	return 0;
}

static struct curl_slist *oci_request_init_headers(struct oci_request *request)
{
	struct curl_slist *curl_headers = NULL;
	char date[HTTP_DATE_FMT_SIZE];
	char *header;
	int err;
	int i;

	if (http_gmtime(date, sizeof(date)) != 0)
		return NULL;

	err = add_header(&curl_headers, strfmt("Date: %s", date));
	if (err)
		return NULL;

	if (request->config->user_agent) {
		err = add_header(&curl_headers,
				 strfmt("User-Agent: %s",
					request->config->user_agent));
		if (err)
			return NULL;
	}

	if (request->flags & OCI_REQUEST_IMDS) {
		err = add_header_const(&curl_headers,
				       "Authorization: Bearer Oracle");
	} else {
		header = oci_auth_header(request->config, date,
					 request->method,
					 request->endpoint,
					 request->path,
					 request->content_length,
					 request->content_type,
					 request->content_sha256);
		err = add_header(&curl_headers, header);
	}

	if (err)
		return NULL;

	if (request->content_type) {
		err = add_header(&curl_headers, strfmt("Content-Type: %s",
						       request->content_type));
		if (err)
			return NULL;
	}

	if (request->content_sha256) {
		err = add_header(&curl_headers, strfmt("X-Content-Sha256: %s",
						       request->content_sha256));
		if (err)
			return NULL;
	}

	if (request->headers &&
	    request->headers->request_headers) {
		for (i = 0; i < request->headers->request_hcount; i++) {
			err = add_header_const(&curl_headers,
					       request->headers->request_headers[i]);
			if (err)
				return NULL;
		}
	}

	return curl_headers;
}

/*
 * Return a non-zero value if the request had an error. Return -1 if the
 * request has failed and we got no HTTP status. Return 0 if the HTTP*
 * status code was 2xx, otherwise return the HTTP status.
 */
static int oci_request_error(struct oci_request *request,
			     struct oci_error *error)
{
	const char *error_code, *error_msg;
	json_t *obj_error;
	json_error_t json_error;
	size_t error_buflen;
	int http_status;
	CURLcode status;
	char *error_buf;

	status = curl_easy_getinfo(request->curl, CURLINFO_RESPONSE_CODE,
				   &http_status);

	if (status != CURLE_OK) {
		oci_error(error, "Failed to get HTTP status");
		return -1;
	}

	/*
	 * HTTP status code 2xx report success. We consider that any
	 * other code is an error.
	 */
	if (http_status >= 200 && http_status < 300) {
		/*
		 * Although there is no error, report the http_status in
		 * the error structure.
		 */
		oci_error_request_set(error, status, http_status);
		return 0;
	}

	error_buf = dstore_buffer(&request->body_response);
	error_buflen = request->recv_len;

	if (!error_buf || error_buflen <= 0) {
		oci_error_request_set_verbose(error, status, http_status,
					      NULL, NULL);
		return http_status;
	}

	/*
	 * On error, the OCI API usually returns a JSON Error object
	 * in the request body.
	 */

	obj_error = json_loadb(error_buf, error_buflen,
				  JSON_DECODE_ANY, &json_error);
	if (!obj_error)
		goto unknown_error;

	error_code = json_string_value(json_object_get(obj_error, "code"));
	error_msg = json_string_value(json_object_get(obj_error, "message"));

	if (!error_code && !error_msg) {
		json_decref(obj_error);
		goto unknown_error;
	}

	oci_error_request_set_debug(error, status, http_status, error_code,
				    error_msg, error_buf, error_buflen);

	json_decref(obj_error);

	return http_status;

unknown_error:
	oci_error_request_set_debug(error, status, http_status, "UnknownError",
				    "Fail to decode the response body as an Error resource",
				    error_buf, error_buflen);

	return http_status;
}

/*
 * Implement the same retry sleep delay as the OCI SDK retry strategy
 * (exponential backoff with decorrelated jitter delay).
 */

#define OCI_REQUEST_RETRY_SLEEP_BASE		1	/* seconds */
#define OCI_REQUEST_RETRY_EXP_GROWTH_LOG2	1	/* exp_growth = 2^exp_growth_log2 */
#define OCI_REQUEST_RETRY_JITTER		3
#define OCI_REQUEST_RETRY_SLEEP_MAX		30	/* seconds */
#define OCI_REQUEST_RETRY_ATTEMPT_MAX		8
#define OCI_REQUEST_RETRY_TIME_MAX		600	/* seconds */

static void oci_request_retry_sleep(struct oci_request *request)
{
	int attempt = request->attempt;
	int delay;

	delay = OCI_REQUEST_RETRY_SLEEP_BASE;
	delay *= OCI_REQUEST_RETRY_EXP_GROWTH_LOG2 << (attempt - 1);
	delay += rand() % (OCI_REQUEST_RETRY_JITTER + 1);

	sleep(delay < OCI_REQUEST_RETRY_SLEEP_MAX ?
	      delay : OCI_REQUEST_RETRY_SLEEP_MAX);
}

static bool oci_request_should_retry(struct oci_request *request,
				     CURLcode err_curl, int err_http)
{
	if (request->attempt >= OCI_REQUEST_RETRY_ATTEMPT_MAX)
		return false;

	if (time(NULL) - request->start_time >= OCI_REQUEST_RETRY_TIME_MAX)
		return false;

	/*
	 * CURL reports a lot of errors which are network/connection
	 * related. We should retry after any network/connection error
	 * or timeout because an OCI server can drop a connection at any
	 * point. In particular, connection attempts easily fail when
	 * there are already a lot of connections.
	 *
	 * So we retry on all CURL errors except for errors unrelated
	 * to the network. If it is unclear if a CURL error is network
	 * related then it is better to retry; this will just delay the
	 * failure a little bit.
	 */
	switch (err_curl) {

	case CURLE_OK:
	case CURLE_HTTP_RETURNED_ERROR:
		/* continue to check HTTP error */
		break;

	case CURLE_UNSUPPORTED_PROTOCOL:
	case CURLE_FAILED_INIT:
	case CURLE_URL_MALFORMAT:
	case CURLE_NOT_BUILT_IN:
	case CURLE_OUT_OF_MEMORY:
	case CURLE_FUNCTION_NOT_FOUND:
	case CURLE_ABORTED_BY_CALLBACK:
	case CURLE_UNKNOWN_OPTION:
	case CURLE_SSL_ENGINE_NOTFOUND:
	case CURLE_SSL_ENGINE_SETFAILED:
		/* do not retry */
		return false;

	default:
		/* retry on network/connection error or timeout */
		return true;
	}

	/*
	 * The HTTP error is set to -1 is the CURL perform has failed, or
	 * it was successful but we didn't have an HTTP status.
	 *
	 * We retry using the same strategy as the OCI SDK default retry
	 * strategy i.e. retry on:
	 *  - timeouts and connection errors,
	 *  - HTTP 409 (IncorrectState), 429 (throttles)
	 *  - Any HTTP 5xx (Server Error), except 501 (Not Implemented)
	 */
	if (err_http == -1 ||
	    err_http == 409 || err_http == 429 || err_http == 500 ||
	    err_http >= 502)
		return true;

	return false;
}

static int oci_request_prepare_curl(struct oci_request *request,
				    struct oci_error *error)
{
	char *cert_bundle;
	CURL *curl;
	bool body;

	/*
	 * Note that the Content-Length header is automatically added
	 * by CURL.
	 */
	request->curl_headers = oci_request_init_headers(request);
	if (!request->curl_headers) {
		oci_error(error, "Failed to init headers");
		return -1;
	}

	curl = request->curl;

	/*
	 * Skip all signal handling. When using multiple threads, this
	 * should be set for all handles.
	 */
	curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1L);

	/*
	 * The response to the request doesn't doesn't necessarily have
	 * a body, but it will in case of an error. So we always set the
	 * write function to collect any response.
	 */
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, oci_request_write_cb);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, request);
	curl_easy_setopt(curl, CURLOPT_HTTPHEADER, request->curl_headers);
	curl_easy_setopt(curl, CURLOPT_URL, request->url);

	/* extra certificates to verify the peer with */
	cert_bundle = getenv("REQUESTS_CA_BUNDLE");
	if (cert_bundle)
		curl_easy_setopt(curl, CURLOPT_CAINFO, cert_bundle);

	switch (request->method) {

	case HTTP_METHOD_DELETE:
		curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE");
		body = false;
		break;

	case HTTP_METHOD_GET:
		curl_easy_setopt(curl, CURLOPT_HTTPGET, 1L);
		body = false;
		break;

	case HTTP_METHOD_HEAD:
		curl_easy_setopt(curl, CURLOPT_NOBODY, 1);
		body = false;
		break;

	case HTTP_METHOD_POST:
		curl_easy_setopt(curl, CURLOPT_POST, 1L);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE,
				 request->content_length);
		body = true;
		break;

	case HTTP_METHOD_PUT:
		curl_easy_setopt(curl, CURLOPT_UPLOAD, 1L);
		body = true;
		break;

	default:
		oci_error(error, "Unknown HTTP method");
		return -1;
	}

	if (request->progress) {
#if CURL_AT_LEAST_VERSION(7, 32, 0)
		curl_easy_setopt(curl, CURLOPT_XFERINFOFUNCTION,
				 request->progress);
		curl_easy_setopt(curl, CURLOPT_XFERINFODATA,
				 request->progress_data);
#else
		/*
		 * libcurl libraries prior to 7.32.0 (for example on OL7)
		 * don't have CURLOPT_XFERINFO*, so use the deprecated
		 * CURLOPT_PROGRESS* instead. Note that callback types
		 * are different (curl_off_t instead of double) but as
		 * we only use them for debug, we ignore that.
		 */
		curl_easy_setopt(curl, CURLOPT_PROGRESSFUNCTION,
				 request->progress);
		curl_easy_setopt(curl, CURLOPT_PROGRESSDATA,
				 request->progress_data);
#endif
		curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 0);
	}

	if (body) {
		curl_easy_setopt(curl, CURLOPT_INFILESIZE_LARGE,
				 (curl_off_t)request->content_length);
		curl_easy_setopt(curl, CURLOPT_READFUNCTION,
				 oci_request_read_cb);
		curl_easy_setopt(curl, CURLOPT_READDATA, request);
	}

	if (request->headers &&
	    request->headers->response_headers &&
	    request->headers->response_hcount > 0) {
		curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION,
				 oci_request_header_cb);
		curl_easy_setopt(curl, CURLOPT_HEADERDATA, request);
	}

	if (request->config->debug)
		curl_easy_setopt(curl, CURLOPT_VERBOSE, 1L);

	return 0;
}

static int oci_request_execute(struct oci_request *request,
			       struct oci_config *config,
			       enum http_method method,
			       const char *endpoint, char *path,
			       struct oci_body_cfg request_body_cfg,
			       struct oci_body_cfg response_body_cfg,
			       oci_request_progress_f progress,
			       void *progress_data,
			       struct oci_error *error)
{
	int err_http, err_errno;
	CURLcode err_curl;
	int err;

	request->config = config;
	request->method = method;
	request->endpoint = endpoint;
	request->path = path;
	request->request_body_cfg = request_body_cfg;
	request->response_body_cfg = response_body_cfg;
	request->progress = progress;
	request->progress_data = progress_data;

	err = oci_request_prepare(request, error);
	if (err) {
		oci_error(error, "Failed to init request");
		goto error;
	}

	/*
	 * Perform the request. If the request fails with an authorization
	 * error then retry the request after having refreshed the
	 * configuration. This is to handle the case where a token/cert
	 * expires while the request is being submitted.
	 */
	request->start_time = time(NULL);
	request->attempt = 1;

	for (;;) {

		OCI_DEBUG_REQUEST(request, "BEGIN: %s\n", request->url);

		err = oci_request_prepare_curl(request, error);
		if (err) {
			oci_error(error, "Failed to prepare CURL request");
			goto error;
		}

		errno = 0;
		err_curl = curl_easy_perform(request->curl);
		err_errno = errno;

		if (err_curl == CURLE_OK) {
			err_http = oci_request_error(request, error);
			if (!err_http) {
				OCI_DEBUG_REQUEST(request, "END: No Error\n");
				break;
			}

			OCI_DEBUG_REQUEST(request, "ERROR: HTTP %d\n", err_http);

			/*
			 * If there is an authorization error then refresh
			 * the configuration (because certificate or token
			 * might have expired). In any case, retry because
			 * authorization can fail for other reasons like
			 * late date in the request header.
			 */
			if (err_http == HTTP_STATUS_UNAUTHORIZED &&
			    oci_config_refresh(request->config)) {
				OCI_DEBUG_REQUEST(request, "INFO: Auth refreshed\n");
			}

		} else {
			oci_error(error, curl_easy_strerror(err_curl));
			err_http = -1;
			OCI_DEBUG_REQUEST(request,
					  "ERROR: CURL %d: %s; errno %d: %s\n",
					  err_curl, curl_easy_strerror(err_curl),
					  err_errno, strerror(err_errno));
		}

		if (request->config->debug)
			oci_error_print(error);

		if (!oci_request_should_retry(request, err_curl, err_http)) {
			OCI_DEBUG_REQUEST(request, "END: No Retry\n");
			goto error;
		}

		/* sleep and retry */

		oci_request_runtime_fini(request);
		oci_request_retry_sleep(request);
		request->attempt++;

		/* re-initialize the request runtime */
		err = oci_request_runtime_init(request);
		if (err) {
			oci_error(error, "Failed to re-init request");
			goto error;
		}

		/*
		 * Reset the error i.e. drop the current error, and
		 * re-initialize the error structure so that we will
		 * return with the error (if any) from the most recent
		 * request.
		 */
		oci_error_reset(error);
	}

	return 0;

error:
	oci_request_fini(request);
	return -1;
}

/*
 * Send a 'get' request. The body of the response is stored in the
 * provided (*bufp). If no buffer is provided (*bufp == NULL) then
 * a buffer is dynamically allocated and return to the caller. The
 * caller is then responsible for freeing that buffer.
 */
int oci_request_get(struct oci_config *config,
		    const char *endpoint, char *path,
		    char **bufp, size_t *sizep,
		    struct oci_headers *headers,
		    struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	if (!bufp || !sizep)
		return -1;

	oci_request_set_headers(&request, headers);

	/*
	 * The 'get' request has no body and we use a memory buffer
	 * for the response body.
	 */
	err = oci_request_execute(&request, config,
				  HTTP_METHOD_GET, endpoint, path,
				  OCI_BODY_NONE, OCI_BODY_BUFFER(*bufp, *sizep),
				  NULL, NULL, error);
	if (err)
		return -1;

	/*
	 * If no buffer was provided then return the buffer which was
	 * allocated.
	 */
	if (!*bufp)
		*bufp = dstore_fetch_buffer(&request.body_response);

	/*
	 * Return the number of bytes stored in the buffer.
	 */
	*sizep = request.recv_len;

	oci_request_fini(&request);

	return 0;
}

/*
 * Send a 'get' request. The body of the response is stored in the
 * file associated with the provided file descriptor (fd).
 */
int oci_request_get_to_file(struct oci_config *config,
			    const char *endpoint, char *path,
			    int fd, off_t offset, size_t size,
			    struct oci_headers *headers,
			    struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	if (fd < 0)
		return -1;

	oci_request_set_headers(&request, headers);

	/*
	 * The 'get' request has no body and we store the response
	 * body in the file associated with the provided file
	 * descriptor.
	 */
	err = oci_request_execute(&request, config,
				  HTTP_METHOD_GET, endpoint, path,
				  OCI_BODY_NONE,
				  OCI_BODY_FD(fd, offset, size),
				  NULL, NULL, error);
	if (err)
		return -1;

	oci_request_fini(&request);

	return 0;
}

static int oci_request_post_common(struct oci_config *config,
				   const char *endpoint, char *path,
				   char *buf, size_t size,
				   struct oci_headers *headers,
				   char **responsep, size_t *response_sizep,
				   struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	oci_request_set_flags(&request, OCI_REQUEST_SHA256);
	oci_request_set_content_type(&request, "application/json");
	oci_request_set_headers(&request, headers);

	err = oci_request_execute(&request, config,
				  HTTP_METHOD_POST, endpoint, path,
				  OCI_BODY_BUFFER(buf, size),
				  OCI_BODY_BUFFER_DYNAMIC,
				  NULL, NULL, error);
	if (err)
		return err;

	if (responsep)
		*responsep = dstore_fetch_buffer(&request.body_response);
	if (response_sizep)
		*response_sizep = request.recv_len;

	oci_request_fini(&request);

	return 0;
}

int oci_request_post(struct oci_config *config,
		     const char *endpoint, char *path,
		     char *buf, size_t size,
		     struct oci_headers *headers,
		     struct oci_error *error)
{
	return oci_request_post_common(config, endpoint, path, buf, size,
				       headers, NULL, NULL, error);
}

json_t *oci_request_post_json(struct oci_config *config,
			      const char *endpoint, char *path,
			      char *buf, size_t size,
			      struct oci_headers *headers,
			      struct oci_error *error)
{
	json_error_t json_error;
	size_t response_size;
	json_t *result;
	char *response;
	int err;

	err = oci_request_post_common(config, endpoint, path, buf, size,
				      headers, &response, &response_size,
				      error);
	if (err)
		return NULL;

	OCI_DEBUG(config, "REQUEST RECV:\n%.*s\n\n",
		  (int)response_size, response);

	result = json_loadb(response, response_size, JSON_DECODE_ANY,
			    &json_error);

	free(response);

	return result;
}

int oci_request_put(struct oci_config *config,
		    const char *endpoint, char *path,
		    char *buf, size_t size,
		    struct oci_headers *headers,
		    oci_request_progress_f progress, void *progress_data,
		    struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	oci_request_set_headers(&request, headers);

	err = oci_request_execute(&request, config,
				  HTTP_METHOD_PUT, endpoint, path,
				  OCI_BODY_BUFFER(buf, size),
				  OCI_BODY_BUFFER_DYNAMIC,
				  progress, progress_data, error);
	if (err)
		return -1;

	oci_request_fini(&request);

	return 0;
}

/*
 * Send a 'put' request. The body of the request is retrieved from
 * the provided file descriptor.
 */
int oci_request_put_from_file(struct oci_config *config,
			      const char *endpoint, char *path,
			      int fd, off_t offset, size_t size,
			      struct oci_headers *headers,
			      oci_request_progress_f progress,
			      void *progress_data,
			      struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	oci_request_set_headers(&request, headers);

	/*
	 * The body of the 'put' request is made by reading the file
	 * with the specified fd. And we use a dynamic memory buffer
	 * to store the response body.
	 */
	err = oci_request_execute(&request, config,
				  HTTP_METHOD_PUT, endpoint, path,
				  OCI_BODY_FD(fd, offset, size),
				  OCI_BODY_BUFFER_DYNAMIC,
				  progress, progress_data, error);
	if (err)
		return -1;

	oci_request_fini(&request);

	return 0;
}

json_t *oci_request_get_json(struct oci_config *config,
			     const char *endpoint, char *path,
			     struct oci_headers *headers,
			     struct oci_error *error)
{
	json_error_t json_error;
	char *buf = NULL;
	size_t size = 0;
	json_t *result;
	int err;

	err = oci_request_get(config, endpoint, path, &buf, &size,
			      headers, error);
	if (err)
		return NULL;

	OCI_DEBUG(config, "REQUEST RECV:\n%.*s\n\n", (int)size, buf);

	result = json_loadb(buf, size, JSON_DECODE_ANY, &json_error);

	free(buf);

	return result;
}

int oci_request_delete(struct oci_config *config,
		       const char *endpoint, char *path,
		       struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	/*
	 * The 'delete' request has no body and we use a dynamic memory
	 * buffer to store the response body.
	 */
	err = oci_request_execute(&request, config,
				  HTTP_METHOD_DELETE, endpoint, path,
				  OCI_BODY_NONE, OCI_BODY_BUFFER_DYNAMIC,
				  NULL, NULL, error);
	if (err)
		return -1;

	oci_request_fini(&request);

	return 0;
}

int oci_request_head(struct oci_config *config,
		     const char *endpoint, char *path,
		     struct oci_headers *headers,
		     struct oci_error *error)
{
	struct oci_request request = { 0 };
	int err;

	oci_request_set_headers(&request, headers);

	/*
	 * The 'head' request has no body and we use a dynamic memory
	 * buffer to store the response body.
	 */
	err = oci_request_execute(&request, config,
				  HTTP_METHOD_HEAD, endpoint, path,
				  OCI_BODY_NONE, OCI_BODY_BUFFER_DYNAMIC,
				  NULL, NULL, error);
	if (err)
		return -1;

	oci_request_fini(&request);

	return 0;
}

/*
 * The instance metadata service periodically experiences short periods
 * of downtime for maintenance. Therefore, when you try to access IMDS
 * endpoints, they might be unavailable. As a best practice, implement
 * retry logic when accessing IMDS endpoints. The following strategy is
 * recommended: retry up to three times with a 30 second timeout if you
 * receive a 404, 429, or 5xx response.
 */

int oci_request_imds_get(struct oci_config *config, char *path,
			 char **bufp, size_t *sizep,
			 struct oci_error *error)

{
	struct oci_request request = { 0 };
	int err;

	/*
	 * IMDS requests can be done without an OCI config, but we need
	 * one for processing (to check the debug level) so create an
	 * empty one if none was specified.
	 */
	if (!config)
		config = &((struct oci_config){ 0 });

	oci_request_set_flags(&request, OCI_REQUEST_IMDS);

	/*
	 * The request has no body and we use a dynamic memory buffer
	 * to store the response body.
	 */
	err = oci_request_execute(&request, config,
				  HTTP_METHOD_GET, NULL, path,
				  OCI_BODY_NONE, OCI_BODY_BUFFER_DYNAMIC,
				  NULL, NULL, error);
	if (err)
		return -1;

	/*
	 * If no buffer was provided then return the buffer which was
	 * allocated.
	 */
	if (!*bufp)
		*bufp = dstore_fetch_buffer(&request.body_response);

	/*
	 * Return the number of bytes stored in the buffer.
	 */
	*sizep = request.recv_len;

	oci_request_fini(&request);

	return 0;
}
