/*
 * Copyright (c) 2023, 2024, Oracle and/or its affiliates.
 *
 * Licensed under the Universal Permissive License v 1.0
 * as shown at https://oss.oracle.com/licenses/upl/
 */

#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>

#include "oci.h"
#include "ocifs.h"
#include "utils.h"
#include "string.h"

/*
 * Multipart Upload (MPU) Tunables
 */

/*
 * File size above which multipart upload is used to copy a file
 * to OCI Object Storage. If it is set to 0 then all files will be
 * copied using multipart upload. Note that the OCI documentation
 * recommend to use multipart upload for files > 100MiB.
 */
static size_t ocifs_cloud_mpu_threshold = OCIFS_DEFAULT_MPU_THRESHOLD;

/*
 * Default number of parts used for multipart upload. This number should
 * not be 0, and it is used as long as it can accommodate the part size
 * limit (i.e. 50GiB). If it set to 1 then files smaller than 50GiB are
 * uploaded with a single request, and files larger than 50GiB are split
 * with parts of 50GiB.
 */
static unsigned int ocifs_cloud_mpu_nparts = OCIFS_DEFAULT_MPU_NPARTS;

/*
 * Number of threads used to send multipart upload requests. Multipart
 * upload requests are sent on file flush or close. If it is set to 0
 * then all parts are sent sequentially by the thread which has called
 * flush/close. If it is set to 1 then all parts are sent sequentially
 * by a single thread. This causes part uploads for all files to be
 * serialized by this single thread.
 */
static unsigned int ocifs_cloud_mpu_nthreads = OCIFS_DEFAULT_MPU_NTHREADS;


struct ocifs_cloud_upload_part;

struct ocifs_cloud_upload {
	const char *object_name;
	const char *upload_id;
	struct ocifs_cloud_upload_part *parts;
	int part_count;		/* total number of parts */
	int part_started;	/* number of parts submitted */
	int part_completed;	/* number of parts completed (uploaded or error) */
	int error;
	int fd;
	bool aborted;
	struct ocifs_cloud_upload *next;
};

struct ocifs_cloud_upload_part {
	struct ocifs_cloud_upload *upload;
	int part_number;
	off_t offset;
	size_t size;
	char *etag;
	time_t progress_time;
};

/*
 * We use a single lock between all upload threads because the
 * concurrent access time will be negliable compared to the upload
 * time of each thread.
 */
static pthread_mutex_t ocifs_cloud_upload_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t ocifs_cloud_upload_cv = PTHREAD_COND_INITIALIZER;

/*
 * Upload queue.
 */
static struct ocifs_cloud_upload *ocifs_cloud_upload_qhead;
static struct ocifs_cloud_upload *ocifs_cloud_upload_qtail;

static pthread_t *ocifs_cloud_upload_tid;
static bool ocifs_cloud_upload_thread_active;


static char *prefix_path(const char *path)
{
	return strfmt("%s/", path);
}

static int get_objects_with_prefix(char *prefix, int limit,
				   struct oci_os_list_objects *list_objects,
				   char *start)
{
	struct oci_os_list_objects_param param = {0};

	param.prefix = prefix;
	param.start = start;
	param.limit = limit;

	return oci_os_list_objects(oci_config, list_objects, &param, NULL);
}

/*
 * Check if a prefix exists in cloud. Return 1 if the prefix exists,
 * 0 if the prefix doesn't exist and -1 if there is an error.
 */
int ocifs_cloud_check_prefix(const char *path)
{
	char *object_name;
	int count;

	/*
	 * To check the prefix exists in cloud, check if there is any object
	 * with a name starting with "<path>/".
	 */

	object_name = prefix_path(path);
	if (!object_name)
		return -1;

	count = get_objects_with_prefix(object_name, 1, NULL, NULL);

	free(object_name);

	if (count < 0)
		return -1;

	return count == 0 ? 0 : 1;
}

int ocifs_cloud_check_object(const char *path)
{
	struct oci_os_object_summary object;
	int count;

	count = oci_os_list_object(oci_config, path, &object, NULL);
	if (count < 0)
		return -1;

	oci_os_object_summary_fini(&object);

	return count > 0 ? 1 : 0;
}

/*
 * Check if a prefix exists in the cloud and is used by other entries.
 * Return:
 *   0  - prefix is not in cloud
 *   1  - prefix is in cloud but not used,
 *  >1  - prefix is in cloud and used
 *  -1  - there is an error
 */
int ocifs_cloud_check_prefix_used(const char *path)
{
	struct oci_os_list_objects list_objects = { 0 };
	char *object_name;
	int count;
	int rv;

	/*
	 * Check if there are multiple (i.e. at least 2) objects with
	 * a name starting with "<path>/". If that's the case then there
	 * is directory and it is not empty.
	 */

	object_name = prefix_path(path);
	if (!object_name)
		return -1;

	count = get_objects_with_prefix(object_name, 2, &list_objects, NULL);

	if (count < 0) {

		/* error */
		rv = -1;

	} else if (count == 0) {

		/* no entry with prefix */
		rv = 0;

	} else if (count == 1) {
		/*
		 * If there is a single object then check that this
		 * is effectively the prefix object. Otherwise this
		 * means that the prefix is used although there is
		 * no prefix object.
		 */
		if (strcmp(list_objects.objects[0].name, object_name) == 0)
			rv = 1;
		else
			rv = 2;
	} else {
		/*
		 * We have at least two object with the prefix so the
		 * prefix exists and it is used.
		 */
		rv = 2;
	}
	
	oci_os_list_objects_fini(&list_objects);
	free(object_name);

	return rv;
}

static void ocifs_cloud_getattr_common(const char *path, struct stat *stbuf,
				       mode_t mode, int nlink, size_t size)
{
	struct timespec now = {0};

	/*
	 * Ignore errors if we can't get time. In case, all times will
	 * be set to 0.
	 */
	(void) clock_gettime(CLOCK_REALTIME, &now);

	stbuf->st_mode = mode;
	stbuf->st_nlink = nlink;
	stbuf->st_size = size;

	stbuf->st_uid = getuid();
	stbuf->st_gid = getgid();

	/*
	 * Set all times to the current time.
	 *
	 * This could be improved by getting some time values from
	 * the object. An object has a timeCreated and a timeModified.
	 * timeCreated can be mapped to statx.stx_btime, and timeModified
	 * to st_mtim. But there is no mapping for st_atim and st_ctim.
	 */

	stbuf->st_atim = now ;	/* last access */
	stbuf->st_mtim = now;	/* last modification */
	stbuf->st_ctim = now;	/* last change */
}

static int ocifs_cloud_getattr_prefix(const char *path, struct stat *stbuf)
{
	char *object_name;
	int count;

	/*
	 * Check if any object has a name starting with "<path>/". If
	 * there is then consider that "<path>" is a directory.
	 */

	object_name = prefix_path(path);
	if (!object_name)
		return -1;
	
	count = get_objects_with_prefix(object_name, 1, NULL, NULL);

	free(object_name);

	if (count < 0)
		return -1;

	if (count == 0)
		return 0;

	ocifs_cloud_getattr_common(path, stbuf, S_IFDIR | 0755, 2, 0);

	return 1;
}

static int ocifs_cloud_getattr_object(const char *path, struct stat *stbuf)
{
	struct oci_os_object_summary object;
	int count;

	/*
	 * Check that an object with the name "<path>" exits. If it does
	 * then consider the object as a regular file.
	 */

	count = oci_os_list_object(oci_config, path, &object, NULL);
	if (count < 0)
		return -1;

	if (count == 0)
		return 0;

	ocifs_cloud_getattr_common(path, stbuf, S_IFREG | 0644, 1, object.size);

	oci_os_object_summary_fini(&object);

	return 1;
}

int ocifs_cloud_getattr(const char *path, struct stat *stbuf)
{
	int rv;

	/*
	 * First, check if a prefix exist. If not, check if an object exists.
	 */

	rv = ocifs_cloud_getattr_prefix(path, stbuf);
	if (rv) {
		/* prefix was found or there was an error */
		return rv;
	}
	
	return ocifs_cloud_getattr_object(path, stbuf);
}

int ocifs_cloud_add_prefix(const char *path)
{
	char *object_name;
	int err;

	object_name = prefix_path(path);
	if (!object_name)
		return -1;

	err = oci_os_put_object(oci_config, object_name, NULL, 0, NULL);
	if (err) {
		free(object_name);
		return -1;
	}

	free(object_name);
	return 0;
}

int ocifs_cloud_rm_prefix(const char *path)
{
	char *object_name;
	int err;

	object_name = prefix_path(path);
	if (!object_name)
		return -1;

	err = oci_os_delete_object(oci_config, object_name, NULL);
	if (err) {
		free(object_name);
		return -1;
	}

	free(object_name);
	return 0;
}

int ocifs_cloud_list_objects(const char *path,
			     struct oci_os_list_objects *list_objects,
			     char *start)
{
	struct oci_os_list_objects_param param = {0};
	int count;

	param.start = start;
	param.delimiter = "/";
	param.fields = "size";
	if (*path != '\0') {
		param.prefix = prefix_path(path);
		if (!param.prefix)
			return -1;
	}

	count = oci_os_list_objects(oci_config, list_objects, &param, NULL);
	free(param.prefix);

	return count;
}

/*
 * Read the part (offset, size) of an object and store it in the
 * specified buffer.
 */
ssize_t ocifs_cloud_read(const char *path, char *buf, size_t size, off_t offset)
{
	struct oci_os_get_object_param param = {0};
	int err;

	param.range_start = offset;
	param.range_size = size;

	err = oci_os_get_object(oci_config, path, &buf, &size, &param, NULL);
	if (err)
		return -1;

	return size;
}

/*
 * Read the part (offset, size) of an object and store it in a file at
 * (offset, size).
 */
ssize_t ocifs_cloud_read_to_file(const char *path, int fd, size_t size, off_t offset)
{
	struct oci_os_get_object_param param = {0};
	int err;

	param.range_start = offset;
	param.range_size = size;

	err = oci_os_get_object_to_file(oci_config, path, fd, offset, size,
					&param, NULL);
	if (err)
		return -1;

	return size;
}

ssize_t ocifs_cloud_write(const char *path, char *buf, size_t size, off_t offset)
{
	errno = ENOTSUP;
	return -1;
}

int ocifs_cloud_add_object(const char *path, char *buffer, size_t size)
{
	return oci_os_put_object(oci_config, path, NULL, 0, NULL);
}

/*
 * Add an upload to the tail of the upload queue.
 *
 * This function should be called with the ocifs_cloud_upload_lock held.
 */
static void ocifs_cloud_upload_enqueue(struct ocifs_cloud_upload *upload)
{
	if (ocifs_cloud_upload_qtail)
		ocifs_cloud_upload_qtail->next = upload;
	else
		ocifs_cloud_upload_qhead = upload;

	ocifs_cloud_upload_qtail = upload;
	upload->next = NULL;
}

/*
 * Remove the upload at the head of the upload queue.
 *
 * This function should be called with the ocifs_cloud_upload_lock held.
 */
static void ocifs_cloud_upload_dequeue(void)
{
	struct ocifs_cloud_upload *upload;

	upload = ocifs_cloud_upload_qhead;
	ocifs_cloud_upload_qhead = upload->next;
	upload->next = NULL;

	if (ocifs_cloud_upload_qtail == upload)
		ocifs_cloud_upload_qtail = NULL;
}


static int ocifs_cloud_upload_part_progress(void *data,
					    curl_off_t dltotal,
					    curl_off_t dlnow,
					    curl_off_t ultotal,
					    curl_off_t ulnow)
{
	struct ocifs_cloud_upload_part *part = data;
	time_t now;

	/*
	 * When the upload is aborted then the upload of each part will
	 * eventually fail with error 404 (Not Found) but this can take
	 * a lot of time especially with large parts.
	 *
	 * So if the upload is aborted then return a non-zero value to
	 * immediately abort the upload of this part.
	 */
	if (part->upload->aborted) {
		OCIFS_DEBUG_CLOUD("Upload %s %s part %d aborting\n",
				  part->upload->upload_id,
				  part->upload->object_name,
				  part->part_number);
		return -1;
	}

	if (!OCIFS_DBG_CLOUD)
		return 0;

	now = time(NULL);

	if (now - part->progress_time > 10) {
		OCIFS_DEBUG_CLOUD("Upload %s %s part %d: %"PRId64"/%"PRId64"\n",
				  part->upload->upload_id,
				  part->upload->object_name,
				  part->part_number, ulnow, ultotal);
		part->progress_time = now;
	}

	return 0;
}


static int ocifs_cloud_upload_part(struct ocifs_cloud_upload_part *part)
{
	struct ocifs_cloud_upload *upload = part->upload;
	struct oci_error error = OCI_ERROR_INIT;
	int err;

	OCIFS_DEBUG_CLOUD("Upload %s %s part %d/%d (%jd, %zd)\n",
			  upload->upload_id, upload->object_name,
			  part->part_number, upload->part_count,
			  part->offset, part->size);

	err = oci_os_upload_part_from_file(oci_config, upload->object_name,
					   upload->upload_id, part->part_number,
					   upload->fd, part->offset, part->size,
					   ocifs_cloud_upload_part_progress,
					   part,
					   &part->etag, &error);

	OCIFS_DEBUG_CLOUD("Upload %s %s part %d/%d done (rv=%d)\n",
			  upload->upload_id, upload->object_name,
			  part->part_number, upload->part_count, err);

	if (OCIFS_DBG_CLOUD && err) {
		oci_error_print(&error);
		oci_error_fini(&error);
	}

	return err;
}

/*
 * Upload Thread.
 *
 * An upload thread process the next upload part of an upload. Uploads
 * are retrieved from the upload queue. If the queue is empty then the
 * thread wait for an upload to be available.
 */
static void *ocifs_cloud_upload_thread(void *arg)
{
	struct ocifs_cloud_upload_part *part;
	struct ocifs_cloud_upload *upload;
	int not_started;
	int error;

	mutex_lock(&ocifs_cloud_upload_lock);

	while (ocifs_cloud_upload_thread_active) {

		upload = ocifs_cloud_upload_qhead;
		if (!upload) {
			cv_wait(&ocifs_cloud_upload_cv,
				&ocifs_cloud_upload_lock);
			continue;
		}

		/*
		 * If the upload had an error but there are still parts
		 * to process then mark all remaining parts as started
		 * and completed, and remove this upload from the queue.
		 */
		if (upload->error) {
			not_started = upload->part_count - upload->part_started;
			upload->part_started += not_started;
			upload->part_completed += not_started;
			ocifs_cloud_upload_dequeue();
			/* notify if upload is completed */
			if (upload->part_completed == upload->part_count)
				cv_broadcast(&ocifs_cloud_upload_cv);
			continue;
		}

		/*
		 * When an upload is in the queue then the upload is not
		 * completed and there are still parts to process. So
		 * process the next part.
		 */
		part = &upload->parts[upload->part_started++];

		/*
		 * If we are processing the last part of an upload then
		 * remove that it from the queue.
		 */
		if (upload->part_started == upload->part_count)
			ocifs_cloud_upload_dequeue();

		/*
		 * Upload the part. We release the lock during the part
		 * upload which can be (very) long.
		 */
		mutex_unlock(&ocifs_cloud_upload_lock);
		error = ocifs_cloud_upload_part(part);
		mutex_lock(&ocifs_cloud_upload_lock);

		/*
		 * Process the part upload result after taking the
		 * lock back.
		 */

		upload->part_completed++;
		if (error)
			upload->error++;

		/* notify if the upload is completed or has an error */
		if (upload->part_completed == upload->part_count || error)
			cv_broadcast(&ocifs_cloud_upload_cv);
	}

	mutex_unlock(&ocifs_cloud_upload_lock);

	return NULL;
}

/*
 * Execute an upload request: add the request to the thread upload
 * queues, then wait for completion.
 */
static void ocifs_cloud_upload_execute(struct ocifs_cloud_upload *upload)
{
	int err;
	int i;

	/*
	 * The number of threads for multipart upload is zero then
	 * this means that we should directly do the upload ourself.
	 */
	if (!ocifs_cloud_mpu_nthreads) {
		for (i = 0; i < upload->part_count; i++) {
			err = ocifs_cloud_upload_part(&upload->parts[i]);
			if (err) {
				upload->error++;
				break;
			}
		}
		return;
	}

	mutex_lock(&ocifs_cloud_upload_lock);

	/* add the upload to the upload queue */
	ocifs_cloud_upload_enqueue(upload);

	/* notify that a new upload needs processing */
	cv_broadcast(&ocifs_cloud_upload_cv);

	/*
	 * Wait for the request to be completed. The request is completed
	 * where all part uploads are completeed or there was an error.
	 *
	 * Note that if there was an error then we may still have threads
	 * pending and uploading parts. They will be stopped when the
	 * upload is aborted.
	 */
	while (upload->part_completed < upload->part_count &&
	       upload->error == 0)
		cv_wait(&ocifs_cloud_upload_cv, &ocifs_cloud_upload_lock);

	mutex_unlock(&ocifs_cloud_upload_lock);
}

int ocifs_cloud_add_object_from_file(const char *path, int fd, size_t size)
{
	struct oci_os_commit_mpu_arg commit_arg = { 0 };
	struct ocifs_cloud_upload_part *parts;
	struct oci_os_mpu mpu_upload = { 0 };
	struct ocifs_cloud_upload upload;
	size_t part_size;
	int part_count;
	off_t offset;
	int err;
	int i;

	if (size <= ocifs_cloud_mpu_threshold)
		return oci_os_put_object_from_file(oci_config, path, fd, NULL);

	/*
	 * Compute the number of parts. We use a default number of parts
	 * (ocifs_cloud_mpu_nparts) unless this creates parts which are
	 * too big (i.e. larger than 50GiB).
	 */
	if (size < OCI_MPU_PART_SIZE_MAX * ocifs_cloud_mpu_nparts) {
		part_count = ocifs_cloud_mpu_nparts;
	} else {
		part_count = (size + OCI_MPU_PART_SIZE_MAX - 1) /
			OCI_MPU_PART_SIZE_MAX;
	}

	if (part_count > OCI_MPU_PART_COUNT_MAX)
		return -1;

	/*
	 * Compute the part size by just dividing the file size by the
	 * number of parts. That way, the file size is evenly distributed
	 * among the different parts.
	 *
	 * If there is a remainder then it will be added to the last part.
	 * The remainder doesn't need to be distributed among all parts
	 * because it will be small compared to the part sizes (unless
	 * ocifs_cloud_mpu_threshold and ocifs_cloud_mpu_nparts are not
	 * inappropriately set).
	 */
	part_size = size / part_count;

	parts = calloc(part_count, sizeof(struct ocifs_cloud_upload_part));
	if (!parts)
		return -1;

	/*
	 * Start the multipart upload sequence.
	 */

	err = oci_os_create_multipart_upload(oci_config, path, &mpu_upload, NULL);
	if (err) {
		free(parts);
		return -1;
	}

	/*
	 * Prepare the multipart upload parts.
	 */

	upload.object_name = path;
	upload.upload_id = mpu_upload.upload_id;
	upload.parts = parts;
	upload.part_started = 0;
	upload.part_completed = 0;
	upload.error = 0;
	upload.fd = fd;
	upload.next = NULL;
	upload.aborted = false;

	offset = 0;

	for (i = 0; i < part_count; i++) {

		/*
		 * With the way part count/size is computed, size should
		 * not reach 0. Add a safety check, just in case, to avoid
		 * creating parts with an upload size of 0.
		 */
		if (size == 0)
			break;

		/*
		 * Configure the part with all the remaining size if:
		 *
		 * - This is the last upload part. This happens if there
		 *   is a part size remainder, and we add the remainder
		 *   to the last part.
		 *
		 * - The remaining size is less than the part size. This
		 *   shouldn't happen with the way part size is computed.
		 *   But just in case.
		 *
		 * - The part size is 0. This shouldn't happen unless
		 *   multipart upload is used for files with a small
		 *   size compared to ocifs_cloud_mpu_nparts (due to
		 *   inappropriate settings for ocifs_cloud_mpu_threshold
		 *   and ocifs_cloud_mpu_nparts).
		 */
		if (i == part_count - 1 || size < part_size || part_size == 0)
			part_size = size;

		/* part number starts at 1 */
		parts[i].part_number = i + 1;
		parts[i].offset = offset;
		parts[i].size = part_size;
		parts[i].etag = NULL;
		parts[i].upload = &upload;

		size -= part_size;
		offset += part_size;
	}

	upload.part_count = i;

	/*
	 * Execute the multipart upload sequence.
	 */

	ocifs_cloud_upload_execute(&upload);

	if (upload.error)
		goto abort;

	/*
	 * Commit the multipart upload sequence.
	 */

	err = oci_os_commit_mpu_arg_init(&commit_arg);
	if (err)
		goto abort;

	for (i = 0; i < part_count; i++) {
		err = oci_os_commit_mpu_add_commit(&commit_arg,
						   parts[i].part_number,
						   parts[i].etag);
		if (err)
			goto abort;
	}

	err = oci_os_commit_multipart_upload(oci_config, path, upload.upload_id,
					     &commit_arg, NULL);
	if (err)
		goto abort;

done:
	for (i = 0; i < part_count; i++)
		free(parts[i].etag);

	free(parts);
	oci_os_mpu_fini(&mpu_upload);
	oci_os_commit_mpu_arg_fini(&commit_arg);

	return err;

abort:
	/*
	 * Abort the multipart upload sequence.
	 *
	 * If the OCI API abort function fails then we will have a
	 * stale multipart upload on the OCI server. It can be cleanup
	 * manually using the OCI CLI or the OCI console, and it won't
	 * arm OCIFS.
	 */

	err = oci_os_abort_multipart_upload(oci_config, path, upload.upload_id,
					    NULL);
	if (err) {
		OCIFS_DEBUG_CLOUD("Failed to abort multipart upload %s for %s\n",
				  upload.upload_id, path);
	}

	/*
	 * OCIFS effectively aborts the multipart upload by setting
	 * upload.aborted = true. This causes progress callbacks of
	 * each part to abort their CURL requests and corresponding
	 * thread will finish their upload tasks (and return an error).
	 */

	upload.aborted = true;

	OCIFS_DEBUG_CLOUD("Upload %s %s aborted\n",
			  upload.upload_id, upload.object_name);

	/*
	 * We may still have parts being processed, wait for this
	 * processing to be completed.
	 */
	mutex_lock(&ocifs_cloud_upload_lock);
	while (upload.part_completed < upload.part_count)
		cv_wait(&ocifs_cloud_upload_cv, &ocifs_cloud_upload_lock);
	mutex_unlock(&ocifs_cloud_upload_lock);


	err = -1;
	goto done;
}

/*
 * Delete an OCI Object Storage object.
 *
 * Return value:
 *  0 : object doesn't exist
 *  1 : object exists and was successfully deleted
 * -1 : error
 */
int ocifs_cloud_unlink(const char *path)
{
	struct oci_error error = OCI_ERROR_INIT;
	int http_status;
	int err;

	err = oci_os_delete_object(oci_config, path, &error);
	if (err) {
		/*
		 * DeleteObject returns HTTP status 404 if the object
		 * doesn't exist.
		 */
		http_status = error.http_status;
		oci_error_fini(&error);
		return http_status == 404 ? 0 : -1;
	}

	return 1;
}

static int ocifs_cloud_rename_objects(struct oci_os_object_summary *objects,
				       int objects_count,
				       char *prefix_old, char *prefix_new,
				       int prefix_old_len,
				       bool *prefix_old_object_p)
{
	char *old, *new;
	int count;
	int rv;
	int i;

	count = 0;

	for (i = 0; i < objects_count; i++) {
		old = objects[i].name;
		if (strcmp(old, prefix_old) == 0) {
			*prefix_old_object_p = true;
			continue;
		}

		/*
		 * Create the new name by adding the new prefix to the
		 * the part of the name after the old prefix.
		 */
		new = strfmt("%s%s", prefix_new, old + prefix_old_len);
		if (!new) {
			OCI_ERROR("Failed to rename %s prefix %s to %s\n",
				  old, prefix_old, prefix_new);
			return -1;
		}

		rv = ocifs_cloud_rename(old, new);
		if (rv < 0) {
			OCI_ERROR("Failed to rename %s to %s\n", old, new);
			break;
		}

		free(new);
		count++;
	}

	return count;
}

int ocifs_cloud_rename_prefix(const char *oldpath, const char *newpath)
{
	struct oci_os_list_objects list_objects = { 0 };
	bool prefix_old_object;
	int prefix_old_len;
	char *prefix_old;
	char *prefix_new;
	char *start;
	int count;
	int err;
	int rv;

	prefix_old = prefix_path(oldpath);
	if (!prefix_old)
		return -1;

	prefix_new = prefix_path(newpath);
	if (!prefix_new) {
		free(prefix_old);
		return -1;
	}

	prefix_old_len = strlen(prefix_old);

	/*
	 * Track if an old prefix object exists to rename it last.
	 */
	prefix_old_object = false;

	/*
	 * Rename all objects which names start with the old prefix.
	 */
	count = 0;
	start = NULL;
	{
		list_objects = (struct oci_os_list_objects) { 0 };

		rv = get_objects_with_prefix(prefix_old, 0,
					     &list_objects, start);
		free(start);
		if (rv < 0)
			goto error;

		rv = ocifs_cloud_rename_objects(list_objects.objects,
						list_objects.objects_count,
						prefix_old, prefix_new,
						prefix_old_len,
						&prefix_old_object);
		if (rv < 0)
			goto error;

		count += rv;

		start = list_objects.next_start_with;
		if (start) {
			start = strdup(start);
			if (!start) {
				oci_os_list_objects_fini(&list_objects);
				goto error;
			}
		}

		oci_os_list_objects_fini(&list_objects);

	} while (start);

	/*
	 * Finally, rename the prefix object.
	 */
	if (prefix_old_object) {
		rv = ocifs_cloud_rename(prefix_old, prefix_new);
	} else {
		/*
		 * Create a new prefix object if no old prefix was
		 * present.
		 */
		err = oci_os_put_object(oci_config, prefix_new, NULL, 0,
					NULL);
		rv = err ? -1 : 1;
	}

	free(prefix_old);
	free(prefix_new);

	/*
	 * If we are able to rename all objects starting with the old
	 * prefix then we consider that the rename was successful even
	 * if we failed to rename the prefix object itself.
	 */
	if (count)
		return 1;

	/*
	 * There was no object containing the prefix we are renaming. So
	 * the success of the rename only rely on the renaming of the
	 * prefix object.
	 */

	return rv;

error:
	free(prefix_old);
	free(prefix_new);
	return -1;
}

/*
 * Rename an OCI Object Storage object.
 *
 * Return value:
 *  0 : object doesn't exist
 *  1 : object exists and was successfully renamed
 * -1 : error
 */
int ocifs_cloud_rename(const char *oldpath, const char *newpath)
{
	struct oci_error error = OCI_ERROR_INIT;
	int http_status;
	int err;

	err = oci_os_rename_object(oci_config, oldpath, newpath, &error);
	if (err) {
		/*
		 * RenameObject returns HTTP status 404 if the object
		 * doesn't exist.
		 */
		http_status = error.http_status;
		oci_error_fini(&error);
		return http_status == 404 ? 0 : -1;
	}

	return 1;
}

void ocifs_cloud_fini(void)
{
	int i;

	if (!ocifs_cloud_mpu_nthreads)
		return;

	mutex_lock(&ocifs_cloud_upload_lock);
	ocifs_cloud_upload_thread_active = false;
	cv_broadcast(&ocifs_cloud_upload_cv);
	mutex_unlock(&ocifs_cloud_upload_lock);

	if (!ocifs_cloud_upload_tid)
		return;

	for (i = 0; i < ocifs_cloud_mpu_nthreads; i++) {
		if (!ocifs_cloud_upload_tid[i])
			continue;
		(void) pthread_join(ocifs_cloud_upload_tid[i], NULL);
	}

	free(ocifs_cloud_upload_tid);
}

/*
 * Cloud initialization.
 *
 * If a multipart upload parameter (mpu_*) then the default value from
 * ocifs_cloud_mpu_* is used.
 */
int ocifs_cloud_init(ssize_t mpu_threshold, int mpu_nparts, int mpu_nthreads)
{
	if (mpu_threshold >= 0)
		ocifs_cloud_mpu_threshold = mpu_threshold;

	if (mpu_nparts >= 0 && mpu_nparts <= OCI_MPU_PART_COUNT_MAX)
		ocifs_cloud_mpu_nparts = mpu_nparts;

	if (mpu_nthreads >= 0)
		ocifs_cloud_mpu_nthreads = mpu_nthreads;

	if (ocifs_cloud_mpu_nthreads == 0)
		return 0;

	mpu_nthreads = ocifs_cloud_mpu_nthreads;

	ocifs_cloud_upload_tid = calloc(mpu_nthreads, sizeof(pthread_t));
	if (!ocifs_cloud_upload_tid) {
		OCI_ERROR("Failed to allocated cloud upload tid array\n");
		goto error;
	}

	/*
	 * Multipart upload threads are created in ocifs_cloud_mpu_init()
	 * which is called after ocifs has daemonized.
	 */

	return 0;

error:
	ocifs_cloud_fini();
	return -1;
}

void ocifs_cloud_mpu_init(void)
{
	int nthreads;
	int err;
	int i;

	ocifs_cloud_upload_thread_active = true;

	nthreads = 0;

	for (i = 0; i < ocifs_cloud_mpu_nthreads; i++) {
		err = pthread_create(&ocifs_cloud_upload_tid[nthreads], NULL,
				     ocifs_cloud_upload_thread,
				     (void *)(long)i);
		if (err) {
			OCI_ERROR("Failed to create cloud upload thread %d\n", i);
			continue;
		}

		nthreads++;
	}

	if (nthreads < ocifs_cloud_mpu_nthreads) {
		OCI_ERROR("Number of multipart upload threads reduced to %d instead of %d\n"
			  "Multipart upload performance can be impacted\n",
			  nthreads, ocifs_cloud_mpu_nthreads);
		ocifs_cloud_mpu_nthreads = nthreads;
	}
}
